#include <unistd.h>
#include <stdio.h>
#include <dirent.h>
#include <gtk/gtk.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include "installer.h"
#include "instlib.h"    

gchar *kernel_version;
gchar *selected_harddisk;
gchar *selected_swappart;
gchar *selected_rootpart;
gchar *selected_bootpart;
gint selected_fs_nr;
gint root_num;   

/**
  A nice small function that just executes a command.
  Does block the current GUI, but most of the time this isn't a problem.
  Can also use it in a different thread, like when copying all the files over
  to the rootpartition.
  Note that normally the returntype should be checked, but as not every
  execution is necessarily fatal it's not done often in here

  @param command gchar*; command to be executed

  return FALSE and a debug message if an error occurred, TRUE otherwise
*/

gboolean ExecuteCommand(gchar *command) {
	gchar *outbuffer;
	gint exitstatus;

	debug("Exec: %s\n",command);
	if (g_spawn_command_line_sync(command,&outbuffer,NULL,&exitstatus,NULL) == FALSE) {
		debug("Error occured executing command\n");
		return FALSE;
	}
	free(outbuffer);
	return TRUE;
}

/**
  Copy the version of the kernel into kernel_version
*/

void StoreKernelVersion() {
	gchar *cmdline;
	gchar *outbuffer;
	gint exitstatus;
	cmdline = g_strdup_printf("uname -r");
	debug("Exec: uname -r\n");
	if (g_spawn_command_line_sync(cmdline,&outbuffer,NULL,&exitstatus,NULL) == FALSE) {
		debug("Error occured executing command\n");
	}
	else {
		if (kernel_version == NULL)
			g_free(kernel_version);
		kernel_version = g_strdup(outbuffer);
	}
	g_free(cmdline);
}

/**
   Unmounts all partitions of a harddisk

   @param harddisk gchar*; path of the harddisk (ie. /dev/hda)

   return FALSE if an error occurred, TRUE otherwise
*/

gboolean UnmountHarddiskPartitions(gchar *harddisk) {
	gchar *cmdline;
	gchar *entryname;
	GDir *dir = g_dir_open(PATH_MNT_DIR,0,NULL);
	if (dir == NULL)
		return FALSE;
	while ((entryname = (gchar *)g_dir_read_name(dir)) != NULL) {
		if (strncmp(harddisk,entryname,3) == 0) {
			cmdline = g_strdup_printf("/mnt/%s\n",entryname);
			umount(cmdline);
			g_free(cmdline);
		}
	}
	return TRUE;
}

/**
  Set the selected harddisk

  @param harddisk gchar*; harddisk that is selected
*/

void SaveSelectedHarddisk(gchar *harddisk) {
	if (selected_harddisk != NULL)
		g_free(selected_harddisk);
	selected_harddisk = g_strdup(harddisk);
}

/**
  Set the selected swappartition

  @param swappart gchar*; swap partition that is selected
*/

void SaveSelectedSwap(gchar *swappart) {
	if (selected_swappart != NULL)
		g_free(selected_swappart);
	selected_swappart = g_strdup(swappart);
}

/**
  Set the selected root partition

  @param rootpart gchar*; root partition that is selected
*/

void SaveSelectedRoot(gchar *rootpart) {
	if (selected_rootpart != NULL)
		g_free(selected_rootpart);
	selected_rootpart = g_strdup(rootpart);
}

/**
  Set the selected boot partition

  @param bootpart gchar*; boot partition that is selected
*/

void SaveSelectedBoot(gchar *bootpart) {
	if (selected_bootpart != NULL)
		g_free(selected_bootpart);
	selected_bootpart = g_strdup(bootpart);
}

gchar * GetSelectedHarddisk() { return selected_harddisk; }
gchar * GetSelectedSwap() { return selected_swappart; }
gchar * GetSelectedRoot() { return selected_rootpart; }
gchar * GetSelectedBoot() { return selected_bootpart; }	

/**
  Set the selected default filesystem type

  @param filesystem gchar*; ext2, ext3, reiserfs or xfs
*/

void SaveSelectedFilesystemTypeString(gchar *filesystem) {
	if (strcmp(filesystem,"ext2") == 0) {
		selected_fs_nr = 0;
	}
	if (strcmp(filesystem,"ext3") == 0) {
		selected_fs_nr = 1;
	}
	if (strcmp(filesystem,"reiserfs") == 0) {
		selected_fs_nr = 2;
	}
	if (strcmp(filesystem,"xfs") == 0) {
		selected_fs_nr = 3;
	}
	debug("Error: unknown filesystem, using ext3\n");
	selected_fs_nr = 1;
}

gchar *GetSelectedFilesystemName() {
	switch (selected_fs_nr) {
	case (0):
		return "ext2";
		break;
	case (1):
		return "ext3";
		break;
	case (2):
		return "reiserfs";
		break;
	case (3):
		return "xfs";
		break;
	default:
		return "ext3";
		break;
	}
}

void SaveSelectedFilesystemType(gint selected) {
	selected_fs_nr = selected;
}

/**
  Sets up a number of variables, this should be called first
*/

void InitInstallprocess() {
	expert_part = FALSE;
	boot_selected = FALSE;
	initrd = NULL;
	multiple = FALSE;
	cpio_passthrough = FALSE;
	kernel_version = NULL;
	selected_harddisk = NULL;
	selected_swappart = NULL;
	selected_rootpart = NULL;
	selected_bootpart = NULL;
}

/**
   Free's all (most?) of the global variables if set
*/

void FreeInstallProcess() {
	if (kernel_version != NULL)
		g_free(kernel_version);
	if (selected_harddisk != NULL)
		g_free(selected_harddisk);
	if (selected_swappart != NULL)
		g_free(selected_swappart);
	if (selected_rootpart != NULL)
		g_free(selected_rootpart);
	if (selected_bootpart != NULL)
		g_free(selected_bootpart);
}

/**
   Starts the 'partitioner'nth partitioner in the list of partitioners

   @param harddisk gchar*; harddisk which the partitioner will use
   @param partitioner gint; nth partitioner in list
   @param xfree86 gboolean; if the partitioner is started in X environment or not
*/

void StartPartitioner(gchar *harddisk, gint partitioner, gboolean xfree86) {
	gint i = 0;
	GList *commandlist = GetPartitionerListCommands(harddisk,xfree86);
	GList *ptr = commandlist;
	if (commandlist == NULL) {
		debug("Error getting partitionerlist, no items\n");
		return;
	}
	while(1) {
		if (ptr == NULL) {
			debug("Error, not enough partitioners in list\n");
			return;
		}
		if (i == partitioner)
			break;
		ptr = ptr->next;
		i++;
	}
	ExecuteCommand(ptr->data);
	FreeStringList(commandlist);
	//execl(cmdline,arg1);
}

void StartAltPartitioner(gchar *harddisk,gchar *partitioner,gboolean xfree86) {
	gchar *cmdline;
	gchar *outbuffer;
	gint exit_status;
	if (xfree86) {
		cmdline = g_strdup_printf("xterm -fn 10x20 -e %s /dev/%s",partitioner,harddisk);
	}
	else {
		cmdline = g_strdup_printf("%s /dev/%s",partitioner,harddisk);
	}
	debug("Exec: %s\n",cmdline);
	g_spawn_command_line_sync(cmdline,&outbuffer,NULL,&exit_status,NULL);
	g_free(cmdline);
}

void UseSwap(gboolean use_swap) {
	swapon = use_swap;
}

/**
   Turns the selected partition into a swap partition

   @param data gpointer; unused
*/

gpointer MakeSwap(gpointer data) {
	gchar *cmdline;

	// make the swappartition
	cmdline = g_strdup_printf("mkswap %s",GetSelectedSwap());
	ExecuteCommand(cmdline);
	g_free(cmdline);
	swap_done = TRUE;
}

/**
   Converts the selected partition to use the specified filesystem type
   choose the fs, space for specific commands here
   NOTE: likely borked function...

   @param selected_fs gchar*;
   @param selected_part gchar*;

   return TRUE/FALSE
*/

gboolean MakeFS (gchar *selected_fs, gchar *selected_part) {
	gchar *cmdline;
	
	switch (selected_fs_nr) {
	case (0):
	case (1):
		cmdline = g_strdup_printf("mkfs.%s %s",selected_fs,selected_part);
		break;
	case (2):
//mkfs.reiserfs --help: -q | --quiet quiet work without messages, progress and
//           questions. Useful if run in a script. For use by end users only.
		cmdline = g_strdup_printf("mkfs.%s -f -f -q %s",selected_fs,selected_part);
		break;
	case (3):
		cmdline = g_strdup_printf("mkfs.%s -f %s",selected_fs,selected_part);
		break;
	default:
		cmdline = g_strdup_printf("mkfs.%s %s",selected_fs,selected_part);
		break;	
	}
	ExecuteCommand(cmdline);
	g_free(cmdline);
}

/**
   Sets up the root (and auxilirary) partitions

   @param data gpointer; unused
*/

gpointer MakeRoot(gpointer data) {

  gchar cmdline[256];
  gint i;

  /* umount the partitions for sure and check for duplicating */
  // somewhere should be check is target directory empty or not
  /* another partition can be leaved mounted ex. during previous install */
  if (expert_part == TRUE) {
	for (i = 0; i < MAX_EXPERT_MOUNTS; i++) {
	    if (strcmp(expert_selected_parts[i],"") != 0 && 
		strcmp(expert_selected_parts[i],"none") != 0) {
// Not sure about	&& expert_fstab_only[i] != TRUE) 
		if (strcmp(GetSelectedRoot(),expert_selected_parts[i]) != 0){
			sprintf(cmdline,"umount %s",expert_selected_parts[i]);
			ExecuteCommand(cmdline);
		}
		else{
		    debug("root partition can't be mounted somewhere else!!!\n");
		    root_done = FALSE;
		    return;
		}
	    }
	}
  }

  sprintf(cmdline,"umount %s",GetSelectedRoot());
  ExecuteCommand(cmdline);
		
  sprintf(cmdline,"umount %s",ROOT_MOUNTPOINT);
  ExecuteCommand(cmdline);

  MakeFS (GetSelectedFilesystemName(),GetSelectedRoot());
  root_done = TRUE;

  // for each partition selected, make a new filesystem on it
  // the type of filesystem is the same as the root filesystem,
  // as most people shouldn't need any weird different partitiontypes
  if (expert_part == TRUE) {
	for (i = 0; i < MAX_EXPERT_MOUNTS; i++) {
		if (strcmp(expert_selected_parts[i],"none") != 0 &&
		    strcmp(expert_selected_parts[i],"") != 0 &&
		    expert_new_filesystem[i] == TRUE) {

			MakeFS (GetSelectedFilesystemName(),expert_selected_parts[i]);
		}
	}	
  }

  debug("Root partition generated\n");
}

/**
   Executes morphix/install.sh script, if it exists, for every minimodule

   return TRUE/FALSE
*/

gboolean ExecuteMinimoduleInstallscripts() {
	int i;
	gchar *cmdline;
	gchar *filename;
	for (i = 0; i < MAX_MINIMODULES; i++) { /* unlimited? */
		gchar *minimoddir = g_strdup_printf("/mnt/mini/mod%d",i);
		if (g_file_test(minimoddir,G_FILE_TEST_IS_DIR) == TRUE) {
			filename = g_strdup_printf("/mnt/mini/mod%d/morphix/install.sh",i);
			if (g_file_test(filename,G_FILE_TEST_EXISTS) == FALSE) {
				cmdline = g_strdup_printf("mkdir -p %s/mnt/mini/",ROOT_MOUNTPOINT);
				ExecuteCommand(cmdline);
				g_free(cmdline);
				// if no file, copy recursively the minimodule
				cmdline = g_strdup_printf("cp -a /mnt/mini/mod%d %s/mnt/mini/",i,ROOT_MOUNTPOINT);
				ExecuteCommand(cmdline);
				g_free(minimoddir);
				g_free(cmdline);
				g_free(filename);
				continue;
			}
			// else
			cmdline = g_strdup_printf("sh %s",filename);
			ExecuteCommand(cmdline);
			g_free(cmdline);
			g_free(filename);
		}
		g_free(minimoddir);
	}
	return TRUE;
}

/* use syscall instead? */
gboolean MountPartition(gchar *partition, gchar *destination) {
	if (!g_file_test(destination,G_FILE_TEST_EXISTS)) {
		mkdir(destination);
		if (!g_file_test(destination,G_FILE_TEST_EXISTS)) {
			debug("Error: Couldn't make directory %s for mounting of parition %s\n",destination,partition);
			return FALSE;
		}
	}
	gchar *cmdline = g_strdup_printf("mount %s %s",partition,destination);
	ExecuteCommand(cmdline);
	g_free(cmdline);
	return TRUE;
}

/**
   Copies the current filesystem to the root/auxilirary partitions
   Also sets up any configuration files which don't depend on
   post-install questions, and basicly is the whole install process

   @param data gpointer; unused

   return TRUE/FALSE
*/

gpointer BuildRoot(gpointer data) {
	gchar mod_buf[256];
	gchar cmdline[256];
	gchar **mod_array;
	gchar *modules;
	gchar *modules_length;
	gint i;
	FILE *file;
	gint modules_len;

	/* Make the mountdir */
	sprintf(cmdline,"mkdir -p %s",ROOT_MOUNTPOINT);
	ExecuteCommand(cmdline);

	/* mount the newly made partition */
	sprintf(cmdline,"mount -t %s %s %s",GetSelectedFilesystemName(),GetSelectedRoot(),ROOT_MOUNTPOINT);
	ExecuteCommand(cmdline);

	/* mount extra partitions, if these have been selected... */
	if (expert_part == TRUE) {
	    for (i = 0; i < MAX_EXPERT_MOUNTS; i++) {
		if (strcmp(expert_selected_parts[i],"") != 0 && 
		    strcmp(expert_selected_parts[i],"none") != 0 &&
		    expert_fstab_only[i] != TRUE) {
		
		    sprintf(cmdline,"%s%s",
			    ROOT_MOUNTPOINT,expert_selected_mounts[i]);
		    //continue operator is not recommended to use in "for" cycle
		    if (mkdir(cmdline,0755) == -1) {
			debug("Error creating directory %s\n",cmdline);
			continue;
		    }
		    // type of newly created filesystem is selected_fs
		    if (expert_new_filesystem[i] != TRUE) 
			sprintf(cmdline,"mount %s %s%s",expert_selected_parts[i],
			        ROOT_MOUNTPOINT,expert_selected_mounts[i]);
		    else sprintf(cmdline,"mount -t %s %s %s%s",GetSelectedFilesystemName(),expert_selected_parts[i],
		    		ROOT_MOUNTPOINT,expert_selected_mounts[i]);				
		    ExecuteCommand(cmdline);
		}
	    }
	}
	if (cpio_passthrough == TRUE) {
		// using cpio-passthrough way

		sprintf(cmdline,"sh -c \"find /bin -depth | cpio -pdmv --quiet %s\"",ROOT_MOUNTPOINT);
		ExecuteCommand(cmdline);		
		sprintf(cmdline,"sh -c \"find /etc -depth | cpio -pdmv --quiet %s\"",ROOT_MOUNTPOINT);
		ExecuteCommand(cmdline);	
		sprintf(cmdline,"sh -c \"find /home -depth | cpio -pdmv --quiet %s\"",ROOT_MOUNTPOINT);
		ExecuteCommand(cmdline);	
		sprintf(cmdline,"sh -c \"find /lib -depth | cpio -pdmv --quiet %s\"",ROOT_MOUNTPOINT);
		ExecuteCommand(cmdline);
		sprintf(cmdline,"sh -c \"find /morphix -depth | cpio -pdmv --quiet %s\"",ROOT_MOUNTPOINT);
		ExecuteCommand(cmdline);
		sprintf(cmdline,"sh -c \"find /sbin -depth | cpio -pdmv --quiet %s\"",ROOT_MOUNTPOINT);
		ExecuteCommand(cmdline);
		sprintf(cmdline,"sh -c \"find /usr -depth | cpio -pdmv --quiet %s\"",ROOT_MOUNTPOINT);
		ExecuteCommand(cmdline);
		sprintf(cmdline,"sh -c \"find /var -depth | cpio -pdmv --quiet %s\"",ROOT_MOUNTPOINT);
		ExecuteCommand(cmdline);
		sprintf(cmdline,"sh -c \"find /MorphixCD/dev -depth | cpio -pdmv --quiet %s\"",ROOT_MOUNTPOINT);
		ExecuteCommand(cmdline);
		sprintf(cmdline,"sh -c \"find /mnt/mini -depth | cpio -pdmv --quiet %s\"",ROOT_MOUNTPOINT);
		ExecuteCommand(cmdline);
		sprintf(cmdline,"sh -c \"find /MorphixCD/boot -depth | cpio -pdmv --quiet %s\"",ROOT_MOUNTPOINT);
		ExecuteCommand(cmdline);
	}
	else {
		sprintf(cmdline,"cp -a /bin /etc /home /lib /morphix /sbin /usr /var %s",ROOT_MOUNTPOINT);
		ExecuteCommand(cmdline);

//		sprintf(cmdline,"cp -a /var/lib %s/var",ROOT_MOUNTPOINT);
//		ExecuteCommand(cmdline);

		/* copy the dev-files */
		sprintf(cmdline,"cp -a /MorphixCD/dev %s/",ROOT_MOUNTPOINT);
		ExecuteCommand(cmdline);
/* copy the normal bootfiles to the partition */
		sprintf(cmdline,"cp -a /MorphixCD/boot %s/",ROOT_MOUNTPOINT);
		ExecuteCommand(cmdline);
	}

	/* make the boot device */
	sprintf(cmdline,"mkdir %s/boot",ROOT_MOUNTPOINT);
	ExecuteCommand(cmdline);

	/* remove the old hosts file */
	sprintf(cmdline,"rm %s/etc/hosts",ROOT_MOUNTPOINT);
	ExecuteCommand(cmdline);

	sprintf(cmdline,"ln -s /dev/scd0 %s/dev/cdrom",ROOT_MOUNTPOINT);
	ExecuteCommand(cmdline);

	/* make cdrom + floppy mountpoints */
	sprintf(cmdline,"mkdir %s/cdrom",ROOT_MOUNTPOINT);
	ExecuteCommand(cmdline);

	sprintf(cmdline,"mkdir %s/cdrom1",ROOT_MOUNTPOINT);
	ExecuteCommand(cmdline);	

	sprintf(cmdline,"mkdir %s/floppy",ROOT_MOUNTPOINT);
	ExecuteCommand(cmdline);

	sprintf(cmdline,"mkdir %s/mnt",ROOT_MOUNTPOINT);
	ExecuteCommand(cmdline);

	ExecuteMinimoduleInstallscripts();

	sprintf(cmdline,"mkdir %s/proc",ROOT_MOUNTPOINT);
	ExecuteCommand(cmdline);

	sprintf(cmdline,"mkdir %s/root",ROOT_MOUNTPOINT);
	ExecuteCommand(cmdline);	
	
	/* make /etc/modules */

	if (!g_file_get_contents(PROC_MODULES,&modules,&modules_len,NULL)) {
		/* error */
	}
	
	/* generate a modules script */
	mod_array = g_strsplit(modules,"\n",0);
	i = 0;
	sprintf(cmdline,"%s/etc/modules",ROOT_MOUNTPOINT);
	file = fopen(cmdline,"w");
	while (mod_array[i][0] != '\0') {
		sscanf(mod_array[i],"%s",mod_buf);
		// skip any livecd-only modules
		// as most people wouldn't need them anyway,
		// and they might muck up some stuff
		// (well, translucency could)
		if (strncmp(mod_buf,"translucency",strlen("translucency")) != 0 &&
		    strncmp(mod_buf,"cloop",strlen("cloop")) != 0) {
			debug("%s\n",mod_buf);
			fprintf(file,"%s\n",mod_buf);
		}
		i++;
	}

	fclose(file);
	// write a new fstab, with the partitions that have been
	// selected
	sprintf(cmdline,"%s/etc/fstab",ROOT_MOUNTPOINT);

	file = fopen(cmdline,"w");		
	debug("%s \t / %s \t defaults \t 0 \t 1\n",GetSelectedRoot(),GetSelectedFilesystemName());

	fprintf(file,"%s \t / %s \t defaults \t 0 \t 1\n",GetSelectedRoot(),GetSelectedFilesystemName());
	if (swapon) {
		debug("%s \t none \t swap \t sw \t 0 \t 0\n",GetSelectedSwap());
		fprintf(file,"%s \t none \t swap \t sw \t 0 \t 0\n",GetSelectedSwap());

	}

	if (expert_part == TRUE) {
		for (i = 0; i < MAX_EXPERT_MOUNTS; i++) {
			if (strcmp(expert_selected_parts[i],"none") == 0 ||
			    strcmp(expert_selected_parts[i],"") == 0) {
				continue;
			}
			debug("%s \t %s %s \t defaults \t 0 \t 1\n",expert_selected_parts[i],expert_selected_mounts[i],GetSelectedFilesystemName());
			fprintf(file,"%s \t %s %s \t defaults \t 0 \t 1\n",expert_selected_parts[i],expert_selected_mounts[i],GetSelectedFilesystemName());
		}
	}

// TODO: I think is bad solution to hardcode /etc/fstab into installer. I think parts of live-CD fstab should be used instead
	debug("proc  /proc  proc  defaults  0  0\n/dev/fd0  /floppy  vfat  defaults,user,noauto,showexec,umask=022  0  0\n/cdrom /cdrom supermount auto,users,ro,fs=auto,dev=/dev/cdrom  0  0\n");
	fprintf(file,"proc  /proc  proc  defaults  0  0\n/floppy /floppy supermount auto,users,fs=vfat,dev=/dev/fd0  0  0\n/cdrom /cdrom supermount auto,users,ro,fs=auto,dev=/dev/cdrom  0  0\n");
	
	// add any entries in regular fstab that have been added my morphix-base
	// except for any swap-entries, don't need any multiple ones...
	gchar **contents_array;
	gchar *contents_etc;
	gint length;
	if (!g_file_get_contents("/etc/fstab",&contents_etc,&length,NULL)) {
		debug("Can't seem to open /etc/fstab, ignoring\n");
	}
	else {
		contents_array = g_strsplit(contents_etc,"\n",0);
		gint j = 0;
		while (contents_array[j] != NULL) {
			if (strstr(contents_array[j],"Added by morphix-base")) {
				j++;
				if (contents_array[j] != NULL) {
					fprintf(file,"# Added by morphix during install\n");
					gchar temp_str[20];
					sscanf(contents_array[j],"%*s %19s",temp_str);
					if (!strstr(temp_str,"none")) {
						sprintf(cmdline,"mkdir %s/%s",ROOT_MOUNTPOINT,temp_str);
						ExecuteCommand(cmdline); 
					}
					fprintf(file,"%s\n",contents_array[j]);
				}
			}
			j++;
		}
		g_strfreev(contents_array);
		free(contents_etc);
	}
	fclose(file);

	/* umount extra partitions, if these have been selected... */
	if (expert_part == TRUE) {
		for (i = 0; i < MAX_EXPERT_MOUNTS; i++) {
			if (strcmp(expert_selected_parts[i],"none") == 0 || 
			    strcmp(expert_selected_parts[i],"") == 0) {
				continue;
			}
			/* only umount if:
			   - partition has been mounted in the first place
			   - it's not the /boot partition, as it probably
			   still needs a .map file file on it
			   - it's not the /home partition, as that one 
			   still needs to be altered */

			if (expert_fstab_only[i] != TRUE &&
			    strcmp(expert_selected_mounts[i],"/boot") != 0 &&
			    strcmp(expert_selected_mounts[i],"/home") != 0 &&
			    strcmp(expert_selected_mounts[i],"/etc") != 0) {
				
				sprintf(cmdline,"umount %s%s",
					ROOT_MOUNTPOINT,expert_selected_mounts[i]);
				ExecuteCommand(cmdline);
			}
		}
	}
	g_strfreev(mod_array);
	free(modules);
	install_done = TRUE;
}

/**
   Copies the shared libraries for a certain file
   Dunno why this is in here...

   @param src_file gchar*; path to the executable
   @param dest_root gchar*; path to the destination's root directory, which will get the shared libraries of the executable

   return TRUE/FALSE
*/

gboolean CopyLinkedLibraries(gchar *src_file,gchar *dest_root) {
	gint i = 0;
	gint exitstatus;
	gchar *cmdline = NULL;
	gchar *outbuffer = NULL;
	gchar **output_array = NULL;

	if (src_file == NULL || dest_root == NULL)
		return FALSE;

	cmdline = g_strdup_printf("ldd %s",src_file);
	if (cmdline == NULL)
		return FALSE;
	g_spawn_command_line_sync(cmdline,&outbuffer,NULL,&exitstatus,NULL);
	free(cmdline);
	if (outbuffer == NULL)
		return FALSE;
	output_array = g_strsplit(outbuffer,"\n",0);
	free(outbuffer);
	if (output_array == NULL)
		return FALSE;

	/* hack to just make the /lib and /usr/lib dirs of our dest_root
	   we could instead just parse the paths instead... */
	cmdline = g_strdup_printf("mkdir -p %s/lib",dest_root);
	ExecuteCommand(cmdline);
	g_free(cmdline);

	cmdline = g_strdup_printf("mkdir -p %s/usr/lib",dest_root);
	ExecuteCommand(cmdline);
	g_free(cmdline);

	/* parse output of ldd */
	while (output_array[i] != NULL) {
		gchar *path1 = NULL;
		gchar *path2 = NULL;
		gint ret = sscanf(output_array[i],"%as => %as",path1, path2);
		if (ret >= 2) {
			/* parsed a correct line, copy %s into the destination root */
			if (g_file_test(path2,G_FILE_TEST_EXISTS)) {
				cmdline = g_strdup_printf("cp -a %s %s%s",path2,dest_root,path2);
				ExecuteCommand(cmdline);
				g_free(cmdline);
			}
		}
		if (path1 != NULL)
			g_free(path1);
		if (path2 != NULL)
			g_free(path2);
		i++;
	}
	g_strfreev(output_array);
	return TRUE;
}

/**
   Sets up the kernel's initrd. This function should also extract the
   kernel from the boot.img, but as this varies per bootloader we have
   the kernel also in the base module (ie. it already is available as
   /boot/vmlinuz

   Therefore, this function only attempts to set up the initrd.gz for
   the installed' bootloader

*/

void CopyKernel() {
	/* 	
	   Copies the kernel to the right place, and has a lot of 
	   code to make a initrd.gz file that works... */

	gchar modulelist[64][256]; /* FIXME */
	gchar *cmdline;
	gint modulecount = 0;
	struct dirent **ep;
	gint n;
	gchar *scsipath = NULL;

	// path to where the scsi modules are
	scsipath = g_strdup_printf("%s/modules/scsi",LOOP_MOUNTPOINT_1);

	mkdir(LOOP_MOUNTPOINT_1,0744);
	mkdir(LOOP_MOUNTPOINT_2,0744);

	// mount the boot image
	cmdline = g_strdup_printf("mount -t vfat -o loop,ro /cdrom/base/boot.img %s",LOOP_MOUNTPOINT_1);
	ExecuteCommand(cmdline);
	g_free(cmdline);
	
	cmdline = g_strdup_printf("sh -c \"gzip -cd %s/miniroot.gz > %s/miniroot\"",LOOP_MOUNTPOINT_1,ROOT_MOUNTPOINT);
	ExecuteCommand(cmdline);
	g_free(cmdline);
	
	cmdline = g_strdup_printf("umount %s",LOOP_MOUNTPOINT_1);
	ExecuteCommand(cmdline);
	g_free(cmdline);

	cmdline = g_strdup_printf("mount -t ext2 -o loop,ro %s/miniroot %s",ROOT_MOUNTPOINT,LOOP_MOUNTPOINT_1);
	ExecuteCommand(cmdline);
	g_free(cmdline);

	// if we have scsi harddisks, check to see which scsi modules we need
	if (strstr(GetSelectedHarddisk(),"sd") != NULL) {
		n = scandir(scsipath,&ep, NULL,alphasort);
		if (n >= 0) {
			int cnt;
			for (cnt = 0; cnt < n; ++cnt) {
				if (strncmp(ep[cnt]->d_name,".",1) != 0 &&
				    strlen(ep[cnt]->d_name) > 2) {
					strncpy(modulelist[modulecount],ep[cnt]->d_name,strlen(ep[cnt]->d_name) - 2); /* ugly hack to remove the .o from each module. This can be done easier, this can be done better, but it's done */

					modulelist[modulecount++][strlen(ep[cnt]->d_name) -2 ] = '\0';
				}
			}
		}
	}
	if (scsipath != NULL)
		g_free(scsipath);

	gchar *pivotfs = NULL;

	switch (selected_fs_nr) {
	case(0):
		/* ext2 is builtin */
		break;
	case(1):
		/* and so is ext3 */
		break;
	case(2):
		strcpy(modulelist[modulecount++],"reiserfs");
		pivotfs = g_strdup_printf("reiserfs");
		break;
	case(3):
		strcpy(modulelist[modulecount++],"xfs");
		pivotfs = g_strdup_printf("xfs");
		break;
	default:
		break;
	}

	// check to see if an initrd.gz is needed at all...
	if (modulecount > 0) {
		cmdline = g_strdup_printf("dd if=/dev/zero of=%s/ird bs=1k count=2500",ROOT_MOUNTPOINT);
		ExecuteCommand(cmdline);
		g_free(cmdline);
		cmdline = g_strdup_printf("mke2fs -F -m 0 -i 1024 -q %s/ird",ROOT_MOUNTPOINT);
		ExecuteCommand(cmdline);
		g_free(cmdline);
		cmdline = g_strdup_printf("mount -t ext2 -o loop,rw %s/ird %s",ROOT_MOUNTPOINT,LOOP_MOUNTPOINT_2);
		ExecuteCommand(cmdline);
		g_free(cmdline);

		cmdline = g_strdup_printf("cp -a %s/dev %s/etc %s/static %s",
			LOOP_MOUNTPOINT_1,LOOP_MOUNTPOINT_1,LOOP_MOUNTPOINT_1,
			LOOP_MOUNTPOINT_2);
		ExecuteCommand(cmdline);
		g_free(cmdline);
		
		cmdline = g_strdup_printf("mkdir -p %s/proc %s/modules %s/lib %s/mnt",
			LOOP_MOUNTPOINT_2,LOOP_MOUNTPOINT_2,LOOP_MOUNTPOINT_2,
			LOOP_MOUNTPOINT_2);
		ExecuteCommand(cmdline);
		g_free(cmdline);

		FILE *file;

		/* generate a new linuxrc file for loading the modules at 
		   boottime */
		cmdline = g_strdup_printf("%s/linuxrc",LOOP_MOUNTPOINT_2);
		file = fopen(cmdline,"w");
		g_free(cmdline);
		if (file == NULL) {
			debug("Error: Can't build initrd: linuxrc not writeable!\n");
			return;
		}

		fprintf(file,"#!/static/sh\n"); // header to start it all off 

		gint i;
		gint exitstatus;
		gchar *outbuffer;
		gchar **output_array;

		// for each module, check to see if it exists
		// if it does, add it to the initrd.gz and load it at boottime
		// by adding a line in linuxrc
		for (i = 0; i < modulecount; i++) {
			gchar modulepath[256] = ""  ;
			cmdline = g_strdup_printf("modinfo %s",modulelist[i]);
			if (g_spawn_command_line_sync(cmdline,&outbuffer,NULL,&exitstatus,NULL) == FALSE) {
				debug("module not found: %s\n",modulelist[i]);
				continue;
			}
			g_free(cmdline);
			output_array = g_strsplit(outbuffer,"\n",0);
			if (output_array == NULL) {
				g_strfreev(output_array);
				debug("Array could not be split?\n");
				continue;
			}
			sscanf(output_array[0],"%*s %s",modulepath);
			if (strcmp(modulepath,"") == 0) {
				g_strfreev(output_array);
				debug("modulepath empty!?\n");
				continue;
			}
			cmdline = g_strdup_printf("cp %s %s/modules",modulepath,
				LOOP_MOUNTPOINT_2);
			ExecuteCommand(cmdline);
			g_free(cmdline);
			
			// add a line to load the module at boottime
			// this is, after all, the whole idea...
			fprintf(file,"insmod /modules/%s.o\n",modulelist[i]);

			g_strfreev(output_array);
		}
		/* okay, now we have our modules loaded. We should be able 
		   to use pivot_root in order to mount the root fs using a 
		   module not compiled into the kernel */
		if (pivotfs != NULL) {
			if (g_file_test(PIVOT_ROOT_PATH,G_FILE_TEST_EXISTS)) {
				cmdline = g_strdup_printf("mkdir %s/static",LOOP_MOUNTPOINT_2);
				ExecuteCommand(cmdline);
				g_free(cmdline);
				cmdline = g_strdup_printf("mkdir -p %s/mnt/initrd",LOOP_MOUNTPOINT_2);
				ExecuteCommand(cmdline);
				g_free(cmdline);

				cmdline = g_strdup_printf("cp -a %s %s/static/pivot_root",PIVOT_ROOT_PATH,LOOP_MOUNTPOINT_2);
				ExecuteCommand(cmdline);
				g_free(cmdline);
				
				fprintf(file,"mount -t proc proc /proc");
				fprintf(file,"echo 256 > /proc/sys/kernel/real-root-dev");
				fprintf(file,"umount /proc");
				fprintf(file,"mount -t %s %s /mnt",pivotfs,GetSelectedRoot());
				fprintf(file,"/static/pivot_root /mnt /mnt/initrd");
			}
			else {
				debug("Error: Couldn't find pivot_root at %s, is it installed?\n",PIVOT_ROOT_PATH);
			}
			g_free(pivotfs);
		}

		fclose(file); // close linuxrc

		// make it executable
		cmdline = g_strdup_printf("chmod 755 %s/linuxrc",LOOP_MOUNTPOINT_2);
		ExecuteCommand(cmdline);
		g_free(cmdline);

		// clean up time...
		cmdline = g_strdup_printf("umount %s",LOOP_MOUNTPOINT_1);
		ExecuteCommand(cmdline);
		g_free(cmdline);
		
		cmdline = g_strdup_printf("umount %s",LOOP_MOUNTPOINT_2);
		ExecuteCommand(cmdline);
		g_free(cmdline);

		cmdline = g_strdup_printf("rm -rf %s/miniroot",ROOT_MOUNTPOINT);
		ExecuteCommand(cmdline);
		g_free(cmdline);

		// make the initrd.gz file
		cmdline = g_strdup_printf("gzip -9 %s/ird",ROOT_MOUNTPOINT);
		ExecuteCommand(cmdline);
		g_free(cmdline);

		cmdline = g_strdup_printf("mv %s/ird.gz %s/boot/initrd.gz",
			ROOT_MOUNTPOINT, ROOT_MOUNTPOINT);
	    // TODO - if there are errors in initrd creation should be ERROR message
		if ( ExecuteCommand(cmdline) == TRUE )
		    initrd = g_strdup_printf("/boot/initrd.gz");
		g_free(cmdline);
	}

	else {
		cmdline = g_strdup_printf("umount %s",LOOP_MOUNTPOINT_1);
		ExecuteCommand(cmdline);
		g_free(cmdline);
	}

}

/**
   Sets up configuration files that depend on the new hostname of our
   new install

   @param hostname gchar*; name of your new Morphix install
*/

void SetupEtcFiles(gchar *hostname) {
	
	gchar *cmdline;
	FILE *file;

	/* generate a new hosts file, including the name of the box 
	   also has ipv6 stuff in it, so everyone can knock themselfs out */
	cmdline = g_strdup_printf("%s/etc/hosts",ROOT_MOUNTPOINT);
	file = fopen(cmdline,"w");
	if (file == NULL)
		return;
	fprintf(file,"127.0.0.1\t%s\t localhost\n\
	  \
	  # The following lines are desirable for IPv6 capable hosts\n\
	  # (added automatically by netbase upgrade)\n\
	  \
	  ::1     ip6-localhost ip6-loopback\n\
	  fe00::0 ip6-localnet\n\
	  ff00::0 ip6-mcastprefix\n\
	  ff02::1 ip6-allnodes\n\
	  ff02::2 ip6-allrouters\n\
	  ff02::3 ip6-allhosts\n",
	  hostname);
	fclose(file);
	g_free(cmdline);

	// write the hostname to file
	cmdline = g_strdup_printf("%s/etc/hostname",ROOT_MOUNTPOINT);

	file = fopen(cmdline,"w");
	if (file == NULL)
		return;
	fprintf(file,"%s\n",hostname);
	fclose(file);
	g_free(cmdline);

	cmdline = g_strdup_printf("%s/etc/mailname",ROOT_MOUNTPOINT);
	file = fopen(cmdline,"w");
	if (file == NULL)
		return;
	fprintf(file,"%s\n",hostname);
	fclose(file);
	g_free(cmdline);
}

/**
   Change the root password of your new install

   @param password gchar*; plain-text root password
*/

void SetRootPassword(gchar *password) {
	gchar *cmdline;

	cmdline = g_strdup_printf("chroot %s sh -c \"echo \\\"root:%s\\\" | chpasswd\"",ROOT_MOUNTPOINT,password);
	ExecuteCommand(cmdline);
	g_free(cmdline);
}

/**
   Set up the new user, depending on if the name is the same as the default
   "morph" user or not. Note that since 0.4-1c different default usernames
   can be set using the username= bootoption, so this might cause problems

   @param username gchar*; username of the new user
   @param password gchar*; plain-text password of the new user
*/

/* This looks ugly, but it works... sortof :o)*/
void SetUpHomeDir(gchar *username, gchar *password) {
//TODO: User groups should made customizable in future,
    // not hardcoded in executable like now
    const gchar user_groups[64] = "dialout,dip,games,cdrom,video,audio,users";
    gchar *cmdline;
    
    if (strncmp(username,"morph",6) != 0) {

	// add the new group
	cmdline = g_strdup_printf("chroot %s sh -c 'groupadd %s'",ROOT_MOUNTPOINT,username);
	ExecuteCommand(cmdline);
	g_free(cmdline);

	// add the new user
	cmdline = g_strdup_printf("chroot %s sh -c 'useradd -m -d /home/%s -c \"Gnoppix User\" -g %s -s /bin/bash %s'",
		ROOT_MOUNTPOINT, username, username, username);
	ExecuteCommand(cmdline);
	g_free(cmdline);
	// Add the password.
	cmdline = g_strdup_printf("chroot %s sh -c 'echo \"%s:%s\" | chpasswd'",ROOT_MOUNTPOINT,username,password);
	ExecuteCommand(cmdline);
	g_free(cmdline);

	// Add the new user to /etc/sudoers - Even though the file may be deleted later on !
	cmdline = g_strdup_printf("chroot %s sh -c 'echo \"%s ALL=NOPASSWD:ALL\" >> /etc/sudoers'",ROOT_MOUNTPOINT,username);
	ExecuteCommand(cmdline);
	g_free(cmdline);

	// make sure the user's files have the right permissions
	cmdline = g_strdup_printf("chroot %s sh -c 'chown -R %s.%s /home/%s'",ROOT_MOUNTPOINT,username,username,username);
	ExecuteCommand(cmdline);
	g_free(cmdline);
    }
    else {
	// change the morph-user's password
	cmdline = g_strdup_printf("chroot %s sh -c 'echo \"morph:%s\" | chpasswd'",ROOT_MOUNTPOINT,password);
	ExecuteCommand(cmdline);
	g_free(cmdline);
    }
    
    // add user to needed groups
    	cmdline = g_strdup_printf("chroot %s sh -c 'usermod -G %s %s'", ROOT_MOUNTPOINT, user_groups, username);
    	ExecuteCommand(cmdline);
    	g_free(cmdline);
}

/**
   Return the kernel arguments that should be valid for the harddisk install
   This way we also copy over arguments entered at boottime of the livecd,
   possibly solving some installer-related issues...

   return GString* containing the bootoptions. g_string_free this after you're done. Returns NULL if there was an error
*/

GString *GetKernelArgumentsString() {
	int length, i;
	gchar **contents_array = NULL;
	GString *kernelarguments = g_string_new("");
	gchar *contents = NULL;
	if (!g_file_get_contents(PROC_CMDLINE,&contents,&length,NULL)) {
		debug("Warning: Can't get kernel arguments, aborting grub\n");
		return NULL;
	}
	contents_array = g_strsplit(contents," ",0);
	if (contents_array == NULL) {
		g_free(contents);
		debug("Warning: Can't get kernel arguments, aborting grub\n");
		return NULL;
	}
	for (i = 0; ;i++) {
		if (contents_array[i] == NULL)
			break;
		if (strncmp(contents_array[i],"BOOT_IMAGE",10) == 0) {
			continue; /* don't need this one */
		}
		if (strcmp(contents_array[i],"ro") == 0) {
			continue; /* what does this one do? */
		}
		if (strncmp(contents_array[i],"initrd=",strlen("initrd=")) == 0) {
			continue; /* don't want this one either */
		}
		if (strncmp(contents_array[i],"root=",strlen("root=")) == 0) {
			continue; /* don't want this one either */
		}
		if (strncmp(contents_array[i],"init=",strlen("init=")) == 0) {
			continue; /* don't want this one either */
		}
		if (strncmp(contents_array[i],"screen=",strlen("screen=")) == 0) {
			continue; /* don't want this one either */
		}
		if (strncmp(contents_array[i],"depth=",strlen("depth=")) == 0) {
			continue; /* don't want this one either */
		}
		if (strncmp(contents_array[i],"keyboard=",strlen("keyboard=")) == 0) {
			continue; /* don't want this one either */
		}
		if (strncmp(contents_array[i],"xvrefresh=",strlen("xvrefresh")) == 0) {
			continue; /* don't want this one either */
		}
		if (strncmp(contents_array[i],"ramdisk_size=",strlen("ramdisk_size=")) == 0) {
			continue; /* don't want this one either */
		}
		if (strncmp(contents_array[i],"lang=",strlen("lang=")) == 0) {
			continue; /* don't want this one either */
		}
		if (strncmp(contents_array[i],"xmodule=",strlen("xmodule=")) == 0) {
			continue; /* don't want this one either */
		}
		/* rest should be relativly harmless */
		g_string_append_printf(kernelarguments,"%s ",contents_array[i]);
	}
	return kernelarguments;
}

/**
   Sets up Grub on the MBR/Root/Boot partition, depending on the parameter
   given.

  @param grub_mode gint;
  1 - use selected_harddisk MBR
  2 - use selected rootpartition 
  3 - don't use lilo
  4 - use selected bootpartition
  
  return TRUE/FALSE
*/

gboolean SetupGrub(gint grub_mode) {
	gchar kernel[256];
	gchar *cmdline = NULL;
	gchar *outbuffer = NULL;
	gint exitstatus;
	GString *kernelarguments = GetKernelArgumentsString();
	FILE *file = NULL;

	if (grub_mode < 1 || grub_mode > 4) {
		debug("Warning: Invalid grub mode %d, aborting\n",grub_mode);
		return FALSE;
	}
	if (grub_mode == 3) {
		/* grub won't be used, return */
		return TRUE;
	}
	sprintf(kernel,"/boot/vmlinuz");
	/* grub-install is called with --no-floppy parameter,
	   otherwise grub-install hangs if there are no floppy drive */
	// TODO: I think is better to make default MBR, not boot partition
	switch (grub_mode) {
	 case (1): // install to mbr of selected device
		cmdline = g_strdup_printf("grub-install --no-floppy --root-directory=%s/ /dev/%s",ROOT_MOUNTPOINT,GetSelectedHarddisk());
		break;
	 case (2): // install to root partition
		cmdline = g_strdup_printf("grub-install --no-floppy --root-directory=%s/ %s",ROOT_MOUNTPOINT,GetSelectedRoot());
		break;
	 default: // install to boot partitoin
		cmdline = g_strdup_printf("grub-install --no-floppy --root-directory=%s/ %s",ROOT_MOUNTPOINT,GetSelectedBoot());
	}	
	ExecuteCommand(cmdline);
	g_free(cmdline);
	// grub should be installed to /boot/grub now
	cmdline = g_strdup_printf("%s/boot/grub",ROOT_MOUNTPOINT);
	if (g_file_test(cmdline,G_FILE_TEST_EXISTS) != TRUE) {
		debug("Warning: Grub wasn't installed correctly it seems, aborting\n");
		return FALSE;
	}
	g_free(cmdline);
	// TODO: We need to get rid of fucking global variables !!!
	if (initrd==NULL) initrd = g_strdup_printf("");
	switch (grub_mode) {
	 case (1): // install to mbr of device
		cmdline = g_strdup_printf("gengrubmenu /dev/%s %s \"%s\" /boot/vmlinuz %s",
					GetSelectedHarddisk(), GetSelectedRoot(),
					kernelarguments->str, initrd);
		break;
	 case (2): // root partition
		cmdline = g_strdup_printf("gengrubmenu %s %s \"%s\" /boot/vmlinuz %s",
					GetSelectedRoot(), GetSelectedRoot(),
					kernelarguments->str, initrd);
		break;
	 default:
		cmdline = g_strdup_printf("gengrubmenu %s %s \"%s\" /boot/vmlinuz %s",
					GetSelectedBoot(), GetSelectedRoot(),
					kernelarguments->str, initrd);
	}
	g_string_free(kernelarguments,TRUE);
	debug("Exec: %s\n",cmdline);
	if (g_spawn_command_line_sync(cmdline,&outbuffer,NULL,&exitstatus,NULL) == FALSE) {
		debug("Error: error occured executing command:\n");
		debug(outbuffer);
		debug("\n");
	}
	if (outbuffer == NULL) {
		debug("Output of gengrubmenu is empty, aborting\n");
		return FALSE;
	}
	g_free(cmdline);
	cmdline = g_strdup_printf("%s%s",ROOT_MOUNTPOINT,GRUB_MENU_LOCATION);
	file = fopen(cmdline,"w");
	debug("%s\n",outbuffer);
	if (file == NULL) {
		debug("Error: grub failed, couldn't write to output file\n");
		g_free(cmdline);
		return FALSE;
	}
	fprintf(file,"%s\n",outbuffer);
	fclose(file);
	g_free(cmdline);
	g_free(outbuffer);
	// grub should be all setup now
	return TRUE;
}

/* lilo has been removed, see CVS if you still want it */

/**
   Set DMA to on for every harddisk. Not sure if we really want this for all
   though...
*/

void ApplyDMA() {
	GString *kp = GetKernelArgumentsString();
	if (strstr(kp->str,"dma ") != NULL) {
		if (strstr(kp->str,"nodma ") == NULL) {
			GList *hdlist = GetHarddiskList(FALSE); // don't want fluffy name
			GList *ptr = hdlist;
			while (ptr != NULL) {
				gchar *cmdline = g_strdup_printf("echo \"/dev/%s {\n\tmult_sect_io = 16\n\twrite_cache = off\n\tdma = on\n}\" >> /etc/default/hdparm",(char *)ptr->data);
				ExecuteCommand(cmdline);
				g_free(cmdline);
				ptr = ptr->next;
			}
			FreeStringList(hdlist);
		}
	}
	g_string_free(kp,TRUE);
}

void SetupMiscPostInstFiles() {
	gchar *cmdline;

	// TODO: detect if DHCP is used and make correct /etc/network/interfaces in this case (or just let them use networkconf instead)
	ApplyDMA();

	// make sure /tmp exists
	cmdline = g_strdup_printf("mkdir %s/tmp",ROOT_MOUNTPOINT);
	ExecuteCommand(cmdline);
	g_free(cmdline);

	/* make /tmp rwx+sticky */
	cmdline = g_strdup_printf("chmod 1777 %s/tmp",ROOT_MOUNTPOINT);
	ExecuteCommand(cmdline);
	g_free(cmdline);

}

/** 
    Execute the mainmodule-dependant /morphix/[pre|post]inst.sh script
    if it exists. You can use this to script around minor
    configs before/after grub is executed. 
*/

void ExecPreInstShFile() {
	gchar *cmdline;

	if (g_file_test("/morphix/preinst.sh",G_FILE_TEST_EXISTS) == TRUE) {
		debug("Info: Executing preinstall script\n");
		cmdline = g_strdup_printf("chroot %s sh /morphix/preinst.sh",ROOT_MOUNTPOINT);
		ExecuteCommand(cmdline);
		g_free(cmdline);
		return;
	}
	if (g_file_test("/morphix/gtkinstaller/preinst.sh",G_FILE_TEST_EXISTS) == TRUE) {
		debug("Warning: Executing DEPRECATED preinstall script, use /morphix/preinst.sh instead!\n");
		cmdline = g_strdup_printf("chroot %s sh /morphix/gtkinstaller/preinst.sh",ROOT_MOUNTPOINT);
		ExecuteCommand(cmdline);
		g_free(cmdline);
		return;
	}
}

void ExecPostInstShFile() {
	gchar *cmdline;
	if (g_file_test("/morphix/postinst.sh",G_FILE_TEST_EXISTS) == TRUE) {
		debug("Info: Executing postinstall script\n");
		cmdline = g_strdup_printf("chroot %s sh /morphix/postinst.sh",ROOT_MOUNTPOINT);
		ExecuteCommand(cmdline);
		g_free(cmdline);
		return;
	}
	if (g_file_test("/morphix/gtkinstaller/postinst.sh",G_FILE_TEST_EXISTS) == TRUE) {
		debug("Warning: Executing DEPRECATED postinstall script, use /morphix/postinst.sh instead!\n");
		cmdline = g_strdup_printf("chroot %s sh /morphix/gtkinstaller/postinst.sh",ROOT_MOUNTPOINT);
		ExecuteCommand(cmdline);
		g_free(cmdline);
	}
}

/**
   Beginnings of a grub bootfloppy builder. Maybe just use grub-floppy
   instead?
*/

void MakeGrubBootdisk() {
	gchar *cmdline;
	
	cmdline = g_strdup_printf("/bin/dd if=/usr/lib/grub/i386-pc/stage1 of=/dev/fd0 bs=\"512\" count=\"1\"");
	ExecuteCommand(cmdline);
	g_free(cmdline);

	cmdline = g_strdup_printf("/bin/dd if=/usr/lib/grub/i386-pc/stage2 of=/dev/fd0 bs=\"512\" seek=\"1\"");
	ExecuteCommand(cmdline);
	g_free(cmdline);
	/* add menu.lst */
}

void UnmountTarget() {
  gchar *cmdline;
  gint i;

// target/miniroot and expert partitions should be unmounted first

  if (expert_part == TRUE) {
	for (i = 0; i < MAX_EXPERT_MOUNTS; i++) {
	    if (strcmp(expert_selected_parts[i],"") != 0 && 
		strcmp(expert_selected_parts[i],"none") != 0) {
			cmdline = g_strdup_printf("umount %s",expert_selected_parts[i]);
			ExecuteCommand(cmdline);
			g_free(cmdline);
	    }
	}
  }

// Don't know where is the best place for umounting miniroot
	cmdline = g_strdup_printf("umount %s/miniroot",ROOT_MOUNTPOINT);
  	ExecuteCommand(cmdline);
	g_free(cmdline);

	debug("Info: Unmounting target\n");
  	cmdline = g_strdup_printf("umount %s",ROOT_MOUNTPOINT);
  	ExecuteCommand(cmdline);
	g_free(cmdline);
}

/**
   Get the list of partitioner commands

   @param harddisk_str gchar*; name of the harddisk (ie. hda)
   @param xfree86 gboolean; if the partitioners should run in an X environment

   return GList* containing all commands for the partitioners available
*/

GList *GetPartitionerListCommands(gchar *harddisk_str,gboolean xfree86) {
	GList *partlistcom = NULL;
	gchar harddisk[4];
	strncpy(harddisk,harddisk_str,3);
	harddisk[3] = '\0';
	if (g_file_test(PM_PATH,G_FILE_TEST_EXISTS)) {
		if (xfree86) {
			partlistcom = g_list_append(partlistcom,g_strdup_printf("%s /dev/%s",PM_PATH,harddisk));
		}
	}
	if (g_file_test(CFDISK_PATH,G_FILE_TEST_EXISTS)) {
		if (xfree86) {
			partlistcom = g_list_append(partlistcom,g_strdup_printf("xterm -fn 10x20 -e %s /dev/%s",CFDISK_PATH,harddisk));
		}
		else {
			partlistcom = g_list_append(partlistcom,g_strdup_printf("%s /dev/%s",CFDISK_PATH,harddisk));
		}
	}
	if (g_file_test(QTPARTED_PATH,G_FILE_TEST_EXISTS)) {
		if (xfree86) {
			partlistcom = g_list_append(partlistcom,g_strdup_printf("%s /dev/%s",QTPARTED_PATH,harddisk));
		}
	}
	if (g_file_test(PARTED_PATH,G_FILE_TEST_EXISTS)) {
		if (xfree86) {
			partlistcom = g_list_append(partlistcom,g_strdup_printf("xterm -fn 10x20 -e %s /dev/%s",PARTED_PATH,harddisk));
		}
		else {
			partlistcom = g_list_append(partlistcom,g_strdup_printf("%s /dev/%s",PARTED_PATH,harddisk));
		}
	}
	return partlistcom;
}

/**
   English (or gettext'd translated) names of all partitioners
   available

   @param xfree86 gboolean; if the partitioners should run in an X environment

   return GList* containing all partitioner names available
*/   

GList *GetPartitionerListNames(gboolean xfree86) {
	GList *partlistnames = NULL;
	if (g_file_test(PM_PATH,G_FILE_TEST_EXISTS)) {
		if (xfree86) {
			partlistnames = g_list_append(partlistnames,g_strdup(_("Partition Morpher - Graphical Partitioner")));
		}
	}
	if (g_file_test(CFDISK_PATH,G_FILE_TEST_EXISTS)) {
		partlistnames = g_list_append(partlistnames,g_strdup(_("CFDisk - commandline partitioner")));
	}
	if (g_file_test(QTPARTED_PATH,G_FILE_TEST_EXISTS)) {
		if (xfree86) {
			partlistnames = g_list_append(partlistnames,g_strdup(_("QTParted - Graphical Partitioner")));
		}
	}
	if (g_file_test(PARTED_PATH,G_FILE_TEST_EXISTS)) {
		partlistnames = g_list_append(partlistnames,g_strdup(_("Parted - advanced commandline partitioner")));
	}
	return partlistnames;
}

/**
   Free's a GList with gchar*-strings in it
*/

void FreeStringList(GList *list) {
	if (list == NULL)
		return;
	GList *ptr = list;
	while (ptr != NULL) {
		g_free(ptr->data);
		ptr = ptr->next;
	}
	g_list_free(list);
}

/**
   @param list GList*; list of items
   @param str gchar*; name of the item you want
   Return the first item's number with a string with the same contents,
   -1 on error
*/

gint GetStringInListNr(GList *list, gchar *str) {
	gint i = 0;
	GList *ptr = list;
	while (ptr != NULL) {
		if (strcmp(ptr->data,str) == 0) {
			return i;
		}
		i++; ptr = ptr->next;
	}
	return -1;
}			

GList *GetNthItem(GList *list,gint nr) {
	GList *ptr = list;
	gint i = 0;
	for(;i < nr; i++) {
		if (ptr == NULL) {
			return NULL;
		}
		ptr = ptr->next;
	}
	return ptr;
}
/** 
    Get the list of harddisks in this piece o' crap

    if name is set, return descriptive name + size
    if not, just return a list of names of the harddisk devices, without the /dev/ 
*/

GList* GetHarddiskList(gboolean name) { 
	gchar *part;
	gchar **part_array;
	gchar *media;
	gchar *model;
	gchar temp_string[256];
	
	gchar media_buffer[256];
	gchar *hdd_buffer = NULL;
	gchar model_buffer[256];

	gint part_len;
	gint media_len;
	gint model_len;
	gint i, j;
	gint hdd_size;

	GList *harddisk_list = NULL;

	// try to get a list of all harddisk partitions
	if (!g_file_get_contents(PROC_PARTITIONS,&part,&part_len,NULL)) {
		debug("Warning: Can't get partition info from harddisk: %s\n",PROC_PARTITIONS);
		return NULL;
	}
	part_array = g_strsplit(part,"\n",0);
	j = 0;
	for (i = 0; ;i++) {
		if (part_array[i] == NULL)
			break;
		if (sscanf(part_array[i],"%*d %*d %d %s",&hdd_size,temp_string) != 0) {
			if (temp_string[3] != '\0' && 
			(strncmp(temp_string,"hd",2) == 0 ||
			 strncmp(temp_string,"md",2) == 0 ||
			 strncmp(temp_string,"sd",2) == 0)) { // only let through disks and ataraid
				continue;
			}
			hdd_size = hdd_size / 1024;
			// hardware raid
			if (strncmp(temp_string,"ataraid/d",9) == 0) {
				if (strlen(temp_string) > 10) { // don't want partitions
					continue;
				}
				if (name) {
					hdd_buffer = g_strdup_printf("%s with %d MiB, Hardware raid",temp_string,hdd_size);
					debug("Info: scsi hdd found: %s\n",hdd_buffer);
					harddisk_list = g_list_append (harddisk_list,hdd_buffer);
				}
				else { /* add plain device name */
					harddisk_list = g_list_append(harddisk_list,g_strdup(temp_string));
				}
			}
			// normal IDE-type harddisks
			if (strncmp(temp_string,"hd",2) == 0 || strncmp(temp_string,"md",2) == 0) {
				sprintf(media_buffer,"/proc/ide/%s/media",
					temp_string);
				g_file_get_contents(media_buffer,&media,&media_len,NULL);

				if (media == NULL) { // Software raid, mdX?
					if (name) {
						hdd_buffer = g_strdup_printf(hdd_buffer,"RAID: %s",temp_string);
						debug("Info: Software RAID found: %s\n",hdd_buffer);
						harddisk_list = g_list_append (harddisk_list,hdd_buffer);
					}
					else {
						harddisk_list = g_list_append(harddisk_list,g_strdup(temp_string));
					}
				}
				// still the usual IDE-type
				else {
					if (strncmp(media,"disk",4) == 0) {
						if (name) {
							sprintf(model_buffer,"/proc/ide/%s/model",temp_string);
							g_file_get_contents(model_buffer,&model,&model_len,NULL);
							model[strlen(model) -1] = '\0';
							hdd_buffer = g_strdup_printf("%s with %d MiB, %s",temp_string,hdd_size,model);
							debug("Info: hdd found: %s\n",hdd_buffer);
							harddisk_list = g_list_append (harddisk_list,hdd_buffer);
						}
						else {
							harddisk_list = g_list_append(harddisk_list,g_strdup(temp_string));
						}
					}
				}

			}
			// ye good ol' SCSI harddisks.
			// these buggers have caused some problems in the past
			// and will probably still cause more problems
			if (strncmp(temp_string,"sd",2) == 0) {
				if (name) {
					hdd_buffer = g_strdup_printf("%s with %d MiB, SCSI",temp_string,hdd_size);
					debug("Info: scsi hdd found: %s\n",hdd_buffer);
					harddisk_list = g_list_append (harddisk_list,hdd_buffer);
				}
				else {
					harddisk_list = g_list_append(harddisk_list,g_strdup(temp_string));
				}
			}
		}
	}
	g_free(part);
	g_strfreev(part_array);

	return harddisk_list;
}

/**
	Get a list of the partitions.
	@param swap gboolean; means only return the swap partitions otherwise return all linux'ed partitions
	@param fluffy gboolean; will put some extra info behind the partition names for easier partition selection

	returned GList should be free'ed when done
	returns NULL when there's something wrong or when the list is empty
	note that swap will automatically add a No Swap entry
*/

GList *GetPartList(gboolean swap,gboolean fluffy) { /* fluffy wuffy names! */
	gchar *outbuffer;
	gchar **part_array;
	gchar *cmdline;
	gchar *swap_buf;
	gchar *root_buf;
	gint exitstatus;
	gchar device[256];
	gchar size_str[256];

	gint size;
	gint i;
	gint id;

	GList *list = NULL;

	// get a list of partitions on the selected harddisk
	cmdline = g_strdup_printf("fdisk -l /dev/%s",GetSelectedHarddisk());
	if (g_spawn_command_line_sync(cmdline,&outbuffer,NULL,&exitstatus,NULL) == FALSE) {
		debug("Error: No fdisk?");
		return NULL;
	}
	g_free(cmdline);
	part_array = g_strsplit(outbuffer,"\n",0);
	g_free(outbuffer);
	root_num = 0;
	i = 6;
	if (part_array[0] == NULL) {
		debug("Fatal: partitions not found for %s!\n",GetSelectedHarddisk());
		return NULL;
	}
	// for every partition, check to see if it's a default Linux type or
	// a swap type partition we can use
	while (part_array[i] != NULL) {
		if (sscanf(part_array[i],"%255s %*d %*d %255s %d Linux",device,size_str,&id) <= 2) {
			if (sscanf(part_array[i],"%255s * %*d %*d %255s %d Linux",device,size_str,&id) <= 2) {
				/* error */
				i++;
				continue;
			}
		}
		debug("Info: Partition %s size: %s\n",device,size_str);
		size = atoi(size_str);

		if (id == 83 && !swap) {
			debug("Info: linux part found: %s\n",device);
			if (fluffy) {
				size = size / 1024;
				root_buf = g_strdup_printf("%s - size: %d MB",device,size);
				list = g_list_append(list,root_buf);
			}
			else {
				list = g_list_append(list,g_strdup(device));
			}
		}
		if (id == 82 && swap) {
			debug("Info: swap part found: %s\n",device);
			if (fluffy) {
				size = size / 1024;
				swap_buf = g_strdup_printf("%s - size: %d MB",device,size);
				list = g_list_append(list,swap_buf);
			}
			else {
				list = g_list_append(list,g_strdup(device));
			}
		}
		i++;
	}
	g_strfreev(part_array);
	// added for those that don't run with swap. 
	// harddisks are slow anyways...
	if (swap) 
		list = g_list_append (list,g_strdup("No Swap"));
		
	return list;
}

/**
   Returns TRUE if you're root. Else piss off.
*/

gboolean CheckRoot() {
	// standard root-check function
	if (geteuid() != 0) {
		error_msg = (char *) malloc(256);
		strcpy(error_msg,"You must be root to install Gnoppix!\n\n");
		return FALSE;
	}
	return TRUE;
}

/**
   Get a list of partitions on the selected harddisk, where we skip
   the selected root partition

   return GList* with the partition list
*/

GList *GetExpertpartitionList() {

	GList *items = NULL;
	GList *rootlist = GetPartList(FALSE,FALSE); 
	GList *ptr = rootlist;
	// have the same list of partitions for each combo-box. 
	// might have to disclude the root partition, but this isn't for 
	// your average user anyway...
	items = g_list_append (items,g_strdup("none"));
	while (ptr != NULL) {
		if (strcmp(ptr->data,GetSelectedRoot()) == 0) {
			ptr = ptr->next;
			continue;
		}		
		items = g_list_append (items,g_strdup(ptr->data));
		ptr = ptr->next;
	}
	FreeStringList(rootlist);
	return items;
}

/**
  Returns the total size of the cdrom used in MB's
  returns -1 if the cdrom couldn't be found
*/

gint GetCdromSize() {
	gchar *cmdline;
	gchar *outbuffer;
	gchar **output_array;
	gint exitstatus;
	gint i = 0;
	gint total_size_mb = 0;

// I think we should care about really size of copiedfiles and use 
// du -s /usr + du -s /lib and so on ... but for some stange reasons 
// /usr on CD is significally bigger - /usr/lib and /usr/X11R6 are duplicated
// Btw, here should displayed somewhat like "Calculating free space..."
// Because, on slower CPU's and CD's this may take more than 2 minutes

// changed back from /usr, as we only want an estimate
	cmdline = g_strdup_printf("du -s /cdrom");
	g_spawn_command_line_sync(cmdline,&outbuffer,NULL,&exitstatus,NULL);
	output_array = g_strsplit(outbuffer,"\n",0);
	free(outbuffer);

	while (output_array[i] != NULL) {
		gchar size[64];
		gint num_finds = 0;
		num_finds = sscanf(output_array[i],"%63s /cdrom",size);
//		if (num_finds == 0) {
			/* fixme: check for multiple cdrom's... */
//		}
		if (num_finds > 0) {
			// found a cdrom device, use it's size
			total_size_mb = (atoi(size) / 1024);
			break;
		}
		i++;
	}
	g_strfreev(output_array);

	debug("Info: total size cdrom: %d MB\n",total_size_mb);
	
	if (total_size_mb > 0) {
		return total_size_mb;
	}
	debug("Error: size of /cdrom could not be found\n");
	return -1;
}

/**
	Returns the total allocated size of the partition, uses fdisk, can 
	be used when unmounted

	@param partition gchar*; name of the partition (ie. hda1)

	returns size, or -1 when you fscked something up
*/

gint GetTotalPartSize(gchar *partition) {
	gchar *outbuffer;
	gchar **part_array;
	gchar *cmdline;
	gint exitstatus;
	gchar device[256];

	gint size;
	gint i;

	// get a list of partitions on the selected harddisk
	cmdline = g_strdup_printf("fdisk -l /dev/%s",GetSelectedHarddisk());
	if (g_spawn_command_line_sync(cmdline,&outbuffer,NULL,&exitstatus,NULL) == FALSE) {
		debug("Error: no fdisk?");
		return -1;
	}
	g_free(cmdline);
	part_array = g_strsplit(outbuffer,"\n",0);
	g_free(outbuffer);
	root_num = 0;
	i = 6;
	if (part_array[0] == NULL) {
		debug("Fatal: partitions not found for %s!\n",GetSelectedHarddisk());
		return -1;
	}
	// for every partition, check to see if it's a default Linux type or
	// a swap type partition we can use
	while (part_array[i] != NULL) {
		if (sscanf(part_array[i],"%255s %*d %*d %d",device,&size) < 2) {
			if (sscanf(part_array[i],"%255s * %*d %*d %d",device,&size) < 2) {
				/* error */
				i++;
				continue;
			}
		}
		if (device != NULL && partition != NULL) {
			if (strncmp(device,partition,strlen(partition)) == 0) {	
				debug("Info: Partition %s size: %d\n",device,size);
				break; /* AWAY FROM THIS LUNACY!!! AAARRRGGG!!! */
			}
		}
		i++;
	}	
	g_strfreev(part_array);
	if (size > 0) {
		return size / 1024;
	}
	return -1;
}

/**
	NOTE: ASSUMES PARTITION IS MOUNTED!

	@param partition gchar*; name of the partition (ie. hda1)

	returns the used size of the partition in MB's
	returns -1 if the partition couldn't be found
*/

gint GetUsedPartSize(gchar *partition) {
	gchar *cmdline;
	gchar *outbuffer;
	gchar **output_array;
	gint exitstatus;
	gint i = 0;
	gint total_size_mb = 0;

	debug("Info: returning size from partition %s\n",partition);

	cmdline = g_strdup_printf("df -m %s",partition);
	g_spawn_command_line_sync(cmdline,&outbuffer,NULL,&exitstatus,NULL);
	free(cmdline);
	output_array = g_strsplit(outbuffer,"\n",0);
	free(outbuffer);
	i++;
	while (output_array[i] != NULL) {
		gchar device[64];
		gchar size[64];
		gint num_finds = 0;
		num_finds = sscanf(output_array[i],"%63s %*s %63s %*s %*s %*s",device,size);
		debug("Info: filtered used size: %s\n",size);
		
		debug("Info: line: %s\n",output_array[i]);
		if (num_finds > 0 && strcmp(device,partition) == 0) {
			// found the partition (hopefully)
			
			total_size_mb = atoi(size);
			break;
		}
		i++;
	}
	g_strfreev(output_array);
	
	debug("Info: Returned size: %d MiB\n",total_size_mb);

	if (total_size_mb > 0) {
		return total_size_mb;
	}
	return -1;
}

/**
  needs the size of the cdrom (installer media), and the size of the root partition (including any extra partitions). Yes, it's an estimate.

  @param cdrom_size gint; total size of the to-be-installed livecd

  will return the percentage copied to the root partition
*/

gint GetCurrentSize(gint cdrom_size) {
	return cdrom_size * 2.8;
}

/**
   Get the current percentage of the install process

   @param cdrom_size gint; size in MBs of the compressed livecd
   @param curr_rootpart_size gint; size in MBs of the root partition (amount already copied)

   return gdouble between 0.0 and 0.99
*/

gdouble GetCurrentPercentage(gint cdrom_size, gint curr_rootpart_size) {
	gdouble perc;
	gdouble unpacked_cdrom_size = GetCurrentSize(cdrom_size); // changed back, as we only need an estimate 
	perc = (gdouble)curr_rootpart_size / unpacked_cdrom_size;

	debug("Info: total size: %f with cdrom_size: %d and rootpartsize: %d\n",perc,cdrom_size,curr_rootpart_size);
	if (perc >= 0.99)
		return 0.99;
	else if (perc < 0)
		return 0.0;
	else 
		return perc;
}
